/**
 * Copyright (C) 2009-2012 enStratus Networks Inc
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.flexiscale.admin;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Locale;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.admin.Offering;
import org.dasein.cloud.admin.Prepayment;
import org.dasein.cloud.admin.PrepaymentSupport;
import org.dasein.cloud.compute.Platform;
import org.dasein.cloud.flexiscale.FlexiscaleProvider;
import org.dasein.cloud.flexiscale.FlexiscaleStub;
import org.dasein.cloud.flexiscale.network.FlexiscaleIP;
import org.dasein.cloud.identity.ServiceAction;
import org.dasein.cloud.network.IpAddress;

import com.flexiant.extility.FlexiScaleSoapBindingStub;
import com.flexiant.extility.ProductComponent;
import com.flexiant.extility.ProductOffer;
import com.flexiant.extility.Server;
import com.flexiant.extility.VDC;

public class FlexiscaleAccounting implements PrepaymentSupport {
	
	/**
	 * Flexiscale available product offerings types
	 * */
	static public final int productOfferTypes = 12;	
	/**
	 * Flexiscale available product offerings types
	 * */	
	/**
	 * Flexiscale requires users to purchase the unites for using the services.
	 * Auto Topup products: After custmoers have purchased one ofAuto Topup products, it will monitor your account unit balance. 
	 * When custmoers'unit balance reaches "Topup Balance", the system will automatically bill  for another top up.
	 * */
	static public final int AutoUnitsProductOfferType = 0;
	static public final	int UnitsProductOfferType = 6;
	
	static public final int RAMProductOfferType = 1;
	static public final int	CPUProductOfferType = 2;
	static public final int DiskProductOfferType = 3;
	static public final	int FirewallProductOfferType = 4;
	static public final	int NetworkProductOfferType = 5;
	static public final	int ImageProductOfferType = 7;
	static public final	int VlanProductOfferType = 8;
	static public final	int DiskIOProductOfferType = 9;
	static public final	int IPSProductOfferType= 10;
	static public final	int ImageDisckProductOfferType= 11;
	
	
	static private final Logger logger = Logger.getLogger(FlexiscaleAccounting.class);
	private FlexiscaleProvider provider = null;
	private FlexiScaleSoapBindingStub service = null;	
	
	FlexiscaleAccounting(FlexiscaleProvider provider) throws CloudException {
		this.provider = provider;
		FlexiscaleStub locator = new FlexiscaleStub (this.provider);
		this.service = locator.getFlexiscaleStub();
	} 

	@Override
	public Offering getOffering(String productOfferId) throws InternalException,CloudException {
		try {
			for (int productOfferType=0; productOfferType < productOfferTypes; productOfferType++){
				ProductOffer[] productOffers = service.listProductOffers(productOfferType);
				if(productOffers == null)
					continue;
				for(ProductOffer productOffer: productOffers){
					if(productOfferId.equals(productOffer.getProduct_offer_id())){
						Offering offering = this.toOffering(productOffer);
						return offering;
					}
				}
			}
			return null;
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
	}
	

	@Override
	public Prepayment getPrepayment(String autoUnitOfferId) throws InternalException,CloudException {
		try {
			ProductOffer[] productOffers = service.listProductOffers(AutoUnitsProductOfferType);
			for(ProductOffer productOffer: productOffers){
				Prepayment prepayment = toPrepayment(productOffer);
				if(prepayment != null)
					return prepayment;
			}
			return null;
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
	}

	@Override
	public String getProviderTermForOffering(Locale locale) {
		return "Product offers";
	}

	@Override
	public String getProviderTermForPrepayment(Locale locale) {
		return "Auto Topups";
	}

	public Collection<Offering> listOfferings(int productOfferType) throws InternalException,CloudException {
		ArrayList<Offering> list = new ArrayList<Offering>();
		try {
			ProductOffer[] productOffers = service.listProductOffers(productOfferType);
			for(ProductOffer productOffer: productOffers){
				Offering offering = this.toOffering(productOffer);
				if(offering != null)
					list.add(offering);
			}
			return list;
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
	}

	@Override
	public Collection<Offering> listOfferings() throws InternalException,CloudException {
		ArrayList<Offering> list = new ArrayList<Offering>();
		for (int productOfferType=0; productOfferType < productOfferTypes; productOfferType++){
			ArrayList<Offering> newlist = (ArrayList<Offering>) this.listOfferings(productOfferType);
			if(newlist != null)
				list.addAll(newlist);
		}
		return list;
	}
	
	@Override
	public Collection<Prepayment> listPrepayments() throws InternalException,CloudException {
		ArrayList<Prepayment> list = new ArrayList<Prepayment>();
		try {
			ProductOffer[] productOffers = service.listProductOffers(AutoUnitsProductOfferType);
			for(ProductOffer productOffer: productOffers){
				Prepayment prepayment = toPrepayment(productOffer);
				if(prepayment != null)
					list.add(prepayment);
			}
			return list;
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
	}

	@Override 
	public String prepay(String offeringId, int count) throws InternalException, CloudException {
		throw new UnsupportedOperationException("The Flexiscale Cloud does not support prepayments.");
	}
	private Offering toOffering(ProductOffer productOffer) throws CloudException{
		if(productOffer == null) return null;		
		Offering dOffering = new Offering();
		
		dOffering.setFixedFee(productOffer.getMinimum_units());
		dOffering.setUsageFee(productOffer.getUnits());
		
		dOffering.setCurrencyCode("unit(EUR)");
		dOffering.setOfferingId(String.valueOf(productOffer.getProduct_offer_id()));
		dOffering.setPlatform(Platform.UNIX);
		dOffering.setSoftware(null);
				
		dOffering.setSize(productOffer.getDescription());		
	/*	String size = "";
		ProductComponent[] pcs = productOffer.getComponents();		
		for(ProductComponent pc : pcs){
			//size = pc.getName() + "("+ pc.getQuantity()+ ")" + " ";
			size = String.valueOf(pc.getQuantity());
		}		
		dOffering.setSize(size);	*/	
		//dOffering.setDataCenterId(dataCenterId);
		//dOffering.setPeriodInDays(periodInDays);		
		return dOffering;
	}
	
	private Prepayment toPrepayment(ProductOffer productOffer) throws CloudException{
		if(productOffer == null) return null;
		Prepayment prepayment = new Prepayment();
		prepayment.setCurrencyCode("EUR");
		prepayment.setFixedFee(productOffer.getInitial_charge());
		prepayment.setUsageFee(productOffer.getRepeat_charge());
		
		prepayment.setPrepaymentId(String.valueOf(productOffer.getProduct_offer_id()));
		prepayment.setPlatform(Platform.UNIX);		
		prepayment.setSize(productOffer.getDescription());	
		prepayment.setSoftware(null);	
	/*  String size = null;
		ProductComponent[] pcs = productOffer.getComponents();		
		for(ProductComponent pc : pcs){
			size = pc.getName() + "("+ pc.getQuantity()+ ")" + " ";
		}		
		dOffering.setSize(size);*/
		//dOffering.setDataCenterId(dataCenterId);
		//dOffering.setPeriodInDays(periodInDays);
		
		return prepayment;
	}

	@Override
	public String[] mapServiceAction(ServiceAction arg0) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public boolean isSubscribed() throws CloudException, InternalException {
		// TODO Auto-generated method stub
		return true;
	}

}
