/**
 * Copyright (C) 2009-2012 enStratus Networks Inc
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.flexiscale.compute;

import java.io.InputStream;
import java.io.OutputStream;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Locale;

import org.apache.log4j.Logger;

import org.dasein.cloud.AsynchronousTask;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.CloudProvider;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.compute.Architecture;
import org.dasein.cloud.compute.MachineImage;
import org.dasein.cloud.compute.MachineImageFormat;
import org.dasein.cloud.compute.MachineImageState;
import org.dasein.cloud.compute.MachineImageSupport;
import org.dasein.cloud.compute.MachineImageType;
import org.dasein.cloud.compute.Platform;
import org.dasein.cloud.flexiscale.FlexiscaleProvider;
import org.dasein.cloud.flexiscale.FlexiscaleStub;
import org.dasein.cloud.identity.ServiceAction;
import org.dasein.util.Jiterator;
import org.dasein.util.JiteratorPopulator;
import org.dasein.util.PopulatorThread;

import com.flexiant.extility.FlexiScaleSoapBindingStub;
import com.flexiant.extility.Foreignkey;
import com.flexiant.extility.ImageCapabilities;
import com.flexiant.extility.ImageTemplate;
import com.flexiant.extility.ImageTemplateMetadata;
import com.flexiant.extility.Server;

public class FlexiscaleImage implements MachineImageSupport {
	static private final Logger logger = Logger.getLogger(FlexiscaleImage.class);	
	private FlexiscaleProvider provider = null;
	private FlexiScaleSoapBindingStub service = null;

	FlexiscaleImage( FlexiscaleProvider provider) throws CloudException{
		this.provider = provider;
		FlexiscaleStub locator = new FlexiscaleStub (this.provider);
		service = locator.getFlexiscaleStub();
	}
	
	/**
	 * Add Foreign key pair for an image
	 * @param imageId: image id
	 * @param key: key name
	 * @param value: value
	 */	
	public void AddForeignkey(String imageId, String key, String value)throws InternalException, CloudException {
		try {
			ImageTemplate imageTemplate = getFlexiscaleImageTemplate(imageId);
			if(imageTemplate != null){
				Foreignkey foreignKey = service.addForeignKey(imageTemplate.getImage_template_uuid(), key, value);
				if(foreignKey == null){
					throw new CloudException("Fails to add Foreignkey for Image Template");
				}
			}
		} catch (NumberFormatException e) {
			logger.error(e.getMessage());
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e.getMessage());
			throw new CloudException(e);
		}		
	}	

	@Override
	public void downloadImage(String imageId, OutputStream ouput)throws CloudException, InternalException {
		//throw new OperationNotSupportedException("Operation Not Supported");		
	}
	
	public ImageTemplate getFlexiscaleImageTemplate(String imageId) throws CloudException,InternalException {
		if( provider.isFlexiscale() ){
			try {
				ImageTemplate[] its = service.listImageTemplates();
				for(ImageTemplate imageTemplate : its) {
					long image_id = imageTemplate.getImage_template_id();
					if(Long.valueOf(imageId) == image_id){
						return imageTemplate;			
					}
				}
			} catch (RemoteException e) {
				logger.error(e.getMessage());
				throw new CloudException(e);			
			}
		}
		return null;
	}

	@Override
	public MachineImage getMachineImage(String imageId) throws CloudException,InternalException {
		if(imageId != null){	
			MachineImage image = new MachineImage();
			ImageTemplate imageTemplate = getFlexiscaleImageTemplate(imageId);
			if(imageTemplate != null){
				image = toMachineImage(imageTemplate);
				if(image != null){
					return image;
				}
			}
		}
		return null;
	}

	@Override
	public String getProviderTermForImage(Locale locale) {
		return "Image";
	}

	@Override
	public boolean hasPublicLibrary() {
		return false;
	}

	
	//The image can only be handled on paused server
	public boolean checkServerReady(String serverId)throws InternalException, CloudException {
		try {
			Server vm = service.getServer(Long.valueOf(serverId));
			if(vm != null){
				/**
				 * (1, 'Error'),(2, 'Running'),(3, 'Starting'),(4, 'Stopping'),(5, 'Stopped'),(6, 'Rebooting'),(7, 'Installing');
				 */	
				switch(vm.getStatus()){
					case 2: return false;
					case 3: return  false;
					case 4: return false; 
					case 5:	return true; 
					case 6:	return false; 
					case 7:	return false; 
					default: return false;//0?		
				}
			}
					
		} catch (NumberFormatException e) {
			logger.equals(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
		return false;
	}
	
	/** Create image from a disk 
	 * @param disk_id
	 * @param name
	 * @param default_username
	 * @param generate_password, true or false
	 * @return  imageId
	 * */	
	public AsynchronousTask<String> imageVirtualMachine(String vmId,String imageName) throws CloudException, InternalException {
		return imageVirtualMachine(vmId,imageName, "root");
	}
	@Override
	public AsynchronousTask<String> imageVirtualMachine(String vmId, String name, String description) throws CloudException, InternalException {
		if(this.checkServerReady(vmId)){			
			long diskId = this.provider.getComputeServices().getVolumeSupport().getPrimaryDiskIdFromServerId(vmId);
			// Do not generate initial password while creating the virtual machine, use root as default username
			return imageVirtualMachine(diskId,name,"root",false);			
		}else{
			
			throw new CloudException("The server need to be paused when imaging");
		}
		
	
	}
	
	/**
	 * Create image from a disk
	 * @param diskID: disk id
	 * @param imageName: image name
	 * @param vmUserName: user name while lauching vm from this image
	 * @param isGenerateInitialPwd: specify if generating the intial password while lauching a vm (true or false)
	 * @return image name
	 */
	
	public AsynchronousTask<String> imageVirtualMachine(long diskID,String imageName, String vmUserName, boolean isGenerateInitialPwd) throws CloudException, InternalException {
	       final AsynchronousTask<String> task = new AsynchronousTask<String>();
	       final long fdiskID = diskID;
	       final String fimageName = imageName;
	       final boolean fisGenerateInitialPwd = isGenerateInitialPwd;	       
	       Thread t = new Thread() { 
	            public void run() {
	                try {
	                    task.completeWithResult(imageVirtualMachine(fdiskID, fimageName, fimageName, fisGenerateInitialPwd,task));
	                }
	                catch( Throwable t ) {
	                    task.complete(t);
	                }
	            }
	        };	        
	        t.setName("Imaging " + fdiskID + " as " + fimageName);
	        t.setDaemon(true);
	        t.start();
	        return task;
	}
	
    private String imageVirtualMachine(long diskID, String imageName, String userName, boolean isGenerateInitialPwd, AsynchronousTask<String> task) throws CloudException, InternalException {
    	try {
			ImageTemplate imageTemplate= service.createImageTemplate(diskID, imageName, userName, isGenerateInitialPwd);
			if(imageTemplate != null)
				return imageTemplate.getImage_template_name();
		} catch (NumberFormatException e) {
			throw new InternalException(e);
		} catch (RemoteException e) {
			throw new CloudException(e);
		}
    	return null;
    }
    /**
     * converting an instance store-backed Windows instance into an AMI
     */
	@Override
	public AsynchronousTask<String> imageVirtualMachineToStorage(String vmId, String name, String description, String directory) throws CloudException,InternalException {
		// TODO Auto-generated method stub
		return null;
	}

	/**
	 * Create image from URL 
	 * @deprecated {@link #installImageFromUpload(String, String, String, String, String, FlexiscaleDiskProduct, String, String)}
	 */
	@Override
	public String installImageFromUpload(MachineImageFormat arg0,InputStream arg1) throws CloudException, InternalException {
		return null;
	}
	
	/**
	 * Create a volume from a uploaded pre-existing disk image 
	 * @param url: Specified Required URL for the Image
	 * @param checksum: (Optional) Specify the MD5 Checksum
	 * @param connUser: (Optional) Specify the authenticated user required to access the image
	 * @param connPass: (Optional) Specify the password required to access the image
	 * @param imageName: Image name
	 * @param diskProduct: Specify the Required Disk ProductOffer (Size)
	 * @param vdcId: Specify the Required VDC ID of which VDC it will be located in
	 * @return
	 * @throws InternalException
	 * @throws CloudException
	 */
	public AsynchronousTask<String> installImageFromUpload(String imageUrl, String checksum, 
			String connUser, String connPass, String imageName, 
			FlexiscaleDiskProduct diskProduct, String vdcId, String initialUserName) throws InternalException, CloudException {
		if(imageUrl == null || diskProduct == null || imageName == null || vdcId == null)
			return null;
		final AsynchronousTask<String> task = new AsynchronousTask<String>();
	    final long fvdcId = Long.parseLong(vdcId);
	    final long diskProductId = diskProduct.getProductId();
	    final String fimageName = imageName;
	    final String fchecksum = checksum;
	    final String fconnUser = connUser;
	    final String fconnPass = connPass;
	    final String fimageUrl = imageUrl;	
	    final String finitialUserName = initialUserName;	
	    final FlexiscaleDisk volumeSupport = this.provider.getComputeServices().getVolumeSupport();
	    
	    Thread t = new Thread() { 
	      public void run() {
	    	  try {
	    		  task.completeWithResult(volumeSupport.waitForCreateVolumeFromUpload(fimageUrl, fchecksum, fconnUser, fconnPass, fimageName, diskProductId,fvdcId, true, finitialUserName,task));
	    	  }catch( Throwable t ) {
	    		  task.complete(t);
	    	  }
	      	}
	      };
	      t.setName("Creating image " + fimageName + "from" + imageUrl);
	      t.setDaemon(true);
	      t.start();
	      return task;			
	}	

	@Override
	public boolean isImageSharedWithPublic(String machineImageId) throws CloudException,InternalException {
	      MachineImage image = getMachineImage(machineImageId);
	      if( image == null ) {
	            return false;
	      	}
	        String p = (String)image.getTag("public");
	        return (p != null && p.equalsIgnoreCase("true"));
	}

	@Override
	public boolean isSubscribed() throws CloudException, InternalException {
		return true;
	}

	@Override
	public Iterable<MachineImage> listMachineImages() throws CloudException,InternalException {
        PopulatorThread<MachineImage> populator;
        provider.hold();
        populator = new PopulatorThread<MachineImage>(new JiteratorPopulator<MachineImage>() {
            public void populate(Jiterator<MachineImage> iterator) throws CloudException, InternalException {
                try {
                    populateImages(null, iterator);
                }
                finally {
                    provider.release();
                }
            }
        });
        populator.populate();
        return populator.getResult();
	}

	@Override
	public Iterable<MachineImage> listMachineImagesOwnedBy(String owner) throws CloudException, InternalException {
        PopulatorThread<MachineImage> populator;
        final String acct = owner;        
        provider.hold();
        populator = new PopulatorThread<MachineImage>(new JiteratorPopulator<MachineImage>() {
            public void populate(Jiterator<MachineImage> iterator) throws CloudException, InternalException {
                try {
                    populateImages(acct, iterator);
                }
                finally {
                    provider.release();
                }
            }
        });
        populator.populate();
        return populator.getResult();
	}
	
	/**
	 * List sharing user Ids
	 */
	@Override
	public Iterable<String> listShares(String forMachineImageId) throws CloudException,InternalException {
		//throw new OperationNotSupportedException("Operation not support");
		return new ArrayList<String>();
	}

	@Override
	public Iterable<MachineImageFormat> listSupportedFormats() throws CloudException, InternalException {
		// Flexiscale support KVM, Xen & VMWare VSphere
		ArrayList<MachineImageFormat> list = new ArrayList<MachineImageFormat>();
		list.add(MachineImageFormat.VMDK);
		list.add(MachineImageFormat.OVF);
		return list;
	}
	
	private void populateImages(String accountNumber, Jiterator<MachineImage> iterator) throws CloudException, InternalException {
		if( provider.isFlexiscale() ){
			try {
				ImageTemplate[] its = service.listImageTemplates();
				for(ImageTemplate it : its) {							
					MachineImage image = toMachineImage(it);
					if(image != null){
						iterator.push(image);
					}
				}
			} catch (RemoteException e) {
				logger.error(e.getMessage());
				throw new CloudException(e);			
			}
		}
	}
    	
	@Override
	public String registerMachineImage(String atStorageLocation) throws CloudException,InternalException {
		throw new OperationNotSupportedException("Operation Not Supported");		
	}

	@Override
	public void remove(String imageId) throws CloudException, InternalException {
		if(imageId == null)
			return;
		try {
			int returnCode = service.deleteImageTemplate(Long.valueOf(imageId));
			if (returnCode != 1){
				throw new CloudException("Failed to delete image " + imageId);
			}
		} catch (NumberFormatException e) {
			logger.error(e.getStackTrace());
			throw new InternalException("The imageId for flexiscale should be a long type number. Details -> "+ e);
		} catch (RemoteException e) {
			logger.error(e.getStackTrace());
			throw new CloudException(e);
		}
	}
	
	/**
	 * Search images(Flexiscale does not provide platform, architecture information)
	 */

	@Override
	public Iterable<MachineImage> searchMachineImages(String keyword, Platform platform, Architecture architecture) throws CloudException,InternalException {
		ArrayList <MachineImage> raw = (ArrayList<MachineImage>) listMachineImages();	
		ArrayList <MachineImage> list = new ArrayList <MachineImage>();
        for( MachineImage image :  raw ) {
        	// No architecture information 
        	/* if( architecture != null && !architecture.equals(image.getArchitecture()) ) {
        		continue;
			}*/
            if( platform != null && !platform.equals(Platform.UNKNOWN) ) {
                Platform mine = image.getPlatform();
                
                if( platform.isWindows() && !mine.isWindows() ) {
                    continue;
                }
                if( platform.isUnix() && !mine.isUnix() ) {
                    continue;
                }
                if( platform.isBsd() && !mine.isBsd() ) {
                    continue;
                }
                if( platform.isLinux() && !mine.isLinux() ) {
                    continue;
                }
                if( platform.equals(Platform.UNIX) ) {
                    if( !mine.isUnix() ) {
                        continue;
                    }
                }
                else if( !platform.equals(mine) ) {
                    continue;
                }
            }
            if( keyword != null && !keyword.equals("") ) {
                keyword = keyword.toLowerCase();
                if( !image.getProviderMachineImageId().toLowerCase().contains(keyword) ) {
                    if( !image.getName().toLowerCase().contains(keyword) ) {
                        if( !image.getDescription().toLowerCase().contains(keyword) ) {
                            continue;
                        }
                    }
                }
            }
            list.add(image);
        }
		return list;
	}
	
	private String getImageUUID(String imageId) throws CloudException, InternalException {
		if(imageId != null){
			ImageTemplate it = this.getFlexiscaleImageTemplate(imageId);
			if(it != null)
				return it.getImage_template_uuid();
		}
		return null;		
	}


	@Override
	public void shareMachineImage(String imageId, String customerUuid, boolean arg2)throws CloudException, InternalException {
		try {
			String imageUuid = this.getImageUUID(imageId);
			int returnCode = service.publishImage(imageUuid, customerUuid);
			if (returnCode != 1){
				throw new CloudException("Failed to share image " + imageId);
			}
		} catch (NumberFormatException e) {
			logger.error(e.getStackTrace());
			throw new InternalException("The imageId for flexiscale should be a long type number. Details -> "+ e);
		} catch (RemoteException e) {
			logger.error(e.getStackTrace());
			throw new CloudException(e);
		}		

	}

	@Override
	public boolean supportsCustomImages() {
		return true;
	}

	@Override
	public boolean supportsImageSharing() {
		return true;
	}

	@Override
	public boolean supportsImageSharingWithPublic() {
		return true;
	}		

	@Override
	public String transfer(CloudProvider fromCloud, String machineImageId)throws CloudException, InternalException {
		throw new OperationNotSupportedException("Does not support");		
	}	
	
	private MachineImage toMachineImage(ImageTemplate imageTemplate) throws CloudException, InternalException{
		MachineImage image = new MachineImage();
		if(imageTemplate == null) return null;
		long imgageId = imageTemplate.getImage_template_id();
		image.setProviderMachineImageId(String.valueOf(imgageId));
		image.setProviderRegionId(provider.getContext().getRegionId());
		image.setType(MachineImageType.STORAGE);
		String name = imageTemplate.getImage_template_name();		
		image.setName(name);
		//Name as description
		image.setDescription(name);
	
		//No Platform Info
		image.setPlatform(Platform.UNKNOWN);

	
		boolean isPublic = true;
		//If the user can not modify the image, then it is public. Otherwise it is private
		if(imageTemplate.getImage_template_canmodify())
		{
			isPublic=false;
		}
		image.addTag("public", String.valueOf(isPublic));
		
		boolean canModify = imageTemplate.getImage_template_canmodify();
		//If user can modify the image, means the user owns this image
		if(canModify){
			image.setProviderOwnerId(provider.getContext().getAccountNumber());
		}
		image.addTag("modify", String.valueOf(canModify));
		
		try {
			ImageCapabilities ic = service.getOtherImageCapabilities(imageTemplate.getImage_template_id());
			if(ic.getCan_create_server()){
				image.setCurrentState(MachineImageState.ACTIVE);			
			}
			
			ImageTemplateMetadata imageMetadata = service.getImageTemplateMetadata(imgageId);
			String systemMetadata = imageMetadata.getSystem_metadata();
			String publicMetadata = imageMetadata.getPublic_metadata();
			String privateMetadata = imageMetadata.getPrivate_metadata();
			
			if( imageMetadata.getSystem_metadata() != null ){
				image.addTag("SystemMetadata",systemMetadata);
			}
			if( imageMetadata.getPublic_metadata() != null ){
				image.addTag("PublicMetadata",publicMetadata);
			}
			if( imageMetadata.getPrivate_metadata() != null ){
				image.addTag("PrivateMetadata",privateMetadata);
			}
			if( imageMetadata.getRestricted_metadata() != null ){
				image.addTag("RestrictedMetadata",imageMetadata.getRestricted_metadata());
			}
			
			//Hardcode as I64
			if(name.toLowerCase().contains("32-bit")  ){
				image.setArchitecture(Architecture.I32);
			}else if(systemMetadata != null){
				if(systemMetadata.toLowerCase().contains("32-bit") || systemMetadata.toLowerCase().contains("32") && systemMetadata.toLowerCase().contains("bit") ){
					image.setArchitecture(Architecture.I32);
				}
			}else{
				image.setArchitecture(Architecture.I64);
			}
			// Set software as empty			
			image.setSoftware("");
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
		return image;		
	}

	@Override
	public String[] mapServiceAction(ServiceAction arg0) {
		// TODO Auto-generated method stub
		return null;
	}
}
