/**
 * Copyright (C) 2009-2012 enStratus Networks Inc
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.flexiscale.compute;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Locale;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.compute.Snapshot;
import org.dasein.cloud.compute.SnapshotState;
import org.dasein.cloud.compute.SnapshotSupport;
import org.dasein.cloud.flexiscale.FlexiscaleProvider;
import org.dasein.cloud.flexiscale.FlexiscaleStub;
import org.dasein.cloud.identity.ServiceAction;

import com.flexiant.extility.DiskSnapshot;
import com.flexiant.extility.FlexiScaleSoapBindingStub;

public class FlexiscaleSnapshot implements SnapshotSupport {
	
	static private final Logger logger = Logger.getLogger(FlexiscaleSnapshot.class);

	private FlexiscaleProvider provider = null;
	private FlexiScaleSoapBindingStub service = null;
	
	FlexiscaleSnapshot(FlexiscaleProvider provider) throws CloudException{
		this.provider = provider;
		FlexiscaleStub locator = new FlexiscaleStub (this.provider);
		service = locator.getFlexiscaleStub();
	} 
	

	@Override
	public String create(String fromVolumeId, String description) throws InternalException, CloudException {
		return create(fromVolumeId,description, 1 );
	}
	
	/**
	 * Create snapshot
	 * @param snapshotOption: specify how to snapshot: 1 will force to create snapshot even if the server is running
	 * Set to 0 will not create if the server is running
	 */
	public String create(String diskId, String snapshotName,int snapshotOption) throws InternalException, CloudException {
		try {
			DiskSnapshot snapshot = service.createDiskSnapshot(Long.valueOf(diskId), snapshotName,snapshotOption);
			if(snapshot != null){
				return String.valueOf(snapshot.getSnapshot_id());
			}else{
				logger.equals("Fails to snapshot ");
			}				
		} catch (NumberFormatException e) {
			throw new InternalException(e);
		} catch (RemoteException e) {
			throw new CloudException(e);
		}
		return null;
	}

	@Override
	public String getProviderTermForSnapshot(Locale locale) {
		return "snapshot";
	}

	@Override
	public Snapshot getSnapshot(String  snapshotId) throws InternalException, CloudException {
		try {
			DiskSnapshot diskSnapshot = service.getDiskSnapshot(Long.valueOf(snapshotId));
			if(diskSnapshot != null){
				Snapshot snapshot = toSnapshot(diskSnapshot);
				return snapshot;
			}
			return null;
		} catch (NumberFormatException e) {
			throw new InternalException(e);
		} catch (RemoteException e) {
			throw new CloudException(e);
		}
	}
	
	@Override
	public boolean isPublic(String snapshotId) throws InternalException, CloudException {
		return false;
	}
	
	@Override
	public boolean isSubscribed() throws InternalException, CloudException {
		return true;
	}

	@Override
	public Iterable<String> listShares(String arg0) throws InternalException,CloudException {
		
		return new ArrayList<String>();
	}

	@Override
	public Iterable<Snapshot> listSnapshots() throws InternalException,CloudException {
		ArrayList<Snapshot> list = new ArrayList<Snapshot>();
		try {
			//Only list the snapshot in currrent virtual data center
			String regionId = this.provider.getRegionId();
			// Go through each disk within current virtual data center
			DiskSnapshot[] diskSnapshots = service.listAllDiskSnapshots(Long.valueOf(regionId));
			if(diskSnapshots != null){				
				for(DiskSnapshot diskSnapshot : diskSnapshots){
					Snapshot snapshot = toSnapshot(diskSnapshot);
					if( snapshot != null ) {
						list.add(snapshot);
					}
					list.add(snapshot);
				}
			}			
			return list;
		} catch (NumberFormatException e) {
			logger.equals(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.equals(e);
			throw new CloudException(e);
		}
	}

	@Override
	public void remove(String snapshotId) throws InternalException, CloudException {
		try {
			int returncode = service.deleteDiskSnapshot(Long.valueOf(snapshotId));
			if(returncode != 1){
				throw new CloudException("Deleting the "+ snapshotId + " was not successful" );
			}
		} catch (NumberFormatException e) {
			logger.equals(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.equals(e);
			throw new CloudException(e);
		}
	}

	@Override
	public void shareSnapshot(String arg0, String arg1, boolean arg2) throws InternalException, CloudException {
		 throw new OperationNotSupportedException("Flexiscale does not support to share snapshot");		
	}

	@Override
	public boolean supportsSnapshotSharing() throws InternalException, CloudException {
		return false;
	}

	@Override
	public boolean supportsSnapshotSharingWithPublic() throws InternalException, CloudException {
		return false;
	}
	
	private Snapshot toSnapshot(DiskSnapshot diskSnapshot) throws InternalException, CloudException{
		Snapshot snapshot = new Snapshot();
		snapshot.setProviderSnapshotId(String.valueOf(diskSnapshot.getSnapshot_id()));
		snapshot.setVolumeId(String.valueOf(diskSnapshot.getDisk_id()));
		snapshot.setRegionId(provider.getContext().getRegionId());	

		//Set as available by default
		snapshot.setCurrentState(SnapshotState.AVAILABLE);
		snapshot.setName(diskSnapshot.getSnapshot_name());	
		snapshot.setDescription(diskSnapshot.getSnapshot_name());
		
		snapshot.setOwner(this.provider.getContext().getAccountNumber());
		snapshot.setSnapshotTimestamp(diskSnapshot.getTimestamp().getTimeInMillis());
		snapshot.setSizeInGb(diskSnapshot.getSnapshot_capacity());

		return snapshot;
	}


	@Override
	public String[] mapServiceAction(ServiceAction arg0) {
		// TODO Auto-generated method stub
		return new String[0];
	}
}
