/**
 * Copyright (C) 2009-2012 enStratus Networks Inc
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.flexiscale.identity;

import java.io.ByteArrayOutputStream;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Locale;

import javax.annotation.Nonnull;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.flexiscale.FlexiscaleProvider;
import org.dasein.cloud.flexiscale.FlexiscaleStub;
import org.dasein.cloud.identity.SSHKeypair;
import org.dasein.cloud.identity.ServiceAction;
import org.dasein.cloud.identity.ShellKeySupport;

import com.flexiant.extility.FlexiScaleSoapBindingStub;
import com.flexiant.extility.SSHKey;
import com.jcraft.jsch.JSch;
import com.jcraft.jsch.KeyPair;

public class Keypairs implements ShellKeySupport {
	
	static private final Logger logger = Logger.getLogger(Keypairs.class);
	private FlexiscaleProvider provider = null;
	private FlexiScaleSoapBindingStub service = null;	
	
	Keypairs(FlexiscaleProvider provider) throws CloudException {
		this.provider = provider;
		FlexiscaleStub locator = new FlexiscaleStub (this.provider);
		this.service = locator.getFlexiscaleStub();
	} 

	@Override
	public SSHKeypair createKeypair(@Nonnull String keyPairName) throws InternalException,CloudException {
		return createKeypair(keyPairName,"root", "rsa", "","");
	}
	
	public SSHKeypair createKeypair(String keyPairName, String username, String algorithm, String passphrase,String comment) throws InternalException,CloudException{
		String privateKeyString = "";
		String publicKeyString = "";
		int type=0;
		if(algorithm.toLowerCase().equals("rsa")){
			type=KeyPair.RSA;
		}else if(algorithm.toLowerCase().equals("dsa")){
			type=KeyPair.DSA;
		}else {
			System.err.println("does not support "+algorithm+" algorithm");
			return null;
		}
	
		JSch jsch=new JSch();
		try{
			KeyPair kpair=KeyPair.genKeyPair(jsch, type);
			kpair.setPassphrase(passphrase);
			ByteArrayOutputStream priout = new ByteArrayOutputStream();
			ByteArrayOutputStream pubout = new ByteArrayOutputStream();
			kpair.writePrivateKey(priout);
			kpair.writePublicKey(pubout, comment);
			 
			privateKeyString = priout.toString();
			publicKeyString = pubout.toString();
			
			SSHKey sshKey = service.addSSHKey(username, publicKeyString, keyPairName);
			if(sshKey != null){
				
                SSHKeypair keypair = new SSHKeypair();
                

                // Flexiscale uses description to keep the name and id of dasein SSHKey
                keypair.setName(sshKey.getDescription());
                
                keypair.setProviderKeypairId(String.valueOf(sshKey.getKey_id())+":"+sshKey.getDescription());

                
                keypair.setProviderOwnerId(provider.getContext().getAccountNumber());
                keypair.setProviderRegionId(provider.getRegionId());               
                keypair.setPublicKey(publicKeyString);
                keypair.setPrivateKey(privateKeyString.getBytes());
                keypair.setFingerprint(kpair.getFingerPrint());
                kpair.dispose();
                return keypair;				
			}		
			
		}catch(Exception e){
			logger.error(e);
			throw new CloudException(e);
		}
		return null;	
	
	}
	
	private String getFlexiscaleKeyIdFromKeyName(String keypairName) throws InternalException, CloudException{
		SSHKeypair key = this.getKeypair(keypairName);
		if(key !=null ){
			return key.getProviderKeypairId().split(":")[0];
		}
		return null;	
	}

	@Override
	public void deleteKeypair(String keypairName) throws InternalException,CloudException {
		try {
			String keypairId = this.getFlexiscaleKeyIdFromKeyName(keypairName);
			if(keypairId == null){
				throw new CloudException("The key pair " + keypairName + " does not exist!!!");				
			}			
			int returnCode= service.removeSSHKey(Long.parseLong(keypairId));
			if(returnCode != 1){
				logger.error("Fails to remove the ssh without explaination");
				throw new CloudException("Fails to remove the ssh without explaination");
			}				
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
	}

	@Override
	public String getFingerprint(String keypairId) throws InternalException,CloudException {
		 throw new OperationNotSupportedException("Flexiscale does not support to get finger print operation");
 	}
	@Override
	public String getProviderTermForKeypair(Locale locale) {
		return "SSH Key";
	}
	
	@Override
	public boolean isSubscribed(){
        return true;
    }

	@Override
	public Collection<SSHKeypair> list() throws InternalException, CloudException {
		ArrayList<SSHKeypair> list = new ArrayList<SSHKeypair>();
		try {
			SSHKey[] sshKeys= service.listSSHKeys();
			for( SSHKey sshKey : sshKeys){				
                SSHKeypair keypair = new SSHKeypair();
               
                // Flexiscale uses description to keep the name and id of dasein SSHKey
                keypair.setName(sshKey.getDescription());
                
                keypair.setProviderKeypairId(String.valueOf(sshKey.getKey_id())+":"+sshKey.getDescription());

                keypair.setProviderOwnerId(provider.getContext().getAccountNumber());
                keypair.setProviderRegionId(provider.getRegionId());
                keypair.setPublicKey(sshKey.getPublic_key());                
                list.add(keypair);              
			}
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}		
		return list;
	}
	@Override
	public String[] mapServiceAction(ServiceAction arg0) {
		// TODO Auto-generated method stub
		return new String[0];
	}
	@Override
	public SSHKeypair getKeypair(String keyPairName) throws InternalException,CloudException {
		// TODO Auto-generated method stub
		ArrayList <SSHKeypair> list = (ArrayList<SSHKeypair>) this.list();
		for(SSHKeypair key : list){
			if(key.getName().equals(keyPairName)){
				return key;
			}			
		}
		return null;
	}
}
