/**
 * Copyright (C) 2009-2012 enStratus Networks Inc
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.flexiscale.network;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Locale;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.flexiscale.FlexiscaleProvider;
import org.dasein.cloud.flexiscale.FlexiscaleStub;
import org.dasein.cloud.identity.ServiceAction;
import org.dasein.cloud.network.AddressType;
import org.dasein.cloud.network.IpAddress;
import org.dasein.cloud.network.IpAddressSupport;
import org.dasein.cloud.network.IpForwardingRule;
import org.dasein.cloud.network.Protocol;

import com.flexiant.extility.FlexiScaleSoapBindingStub;
import com.flexiant.extility.NetworkInterface;
import com.flexiant.extility.Server;
import com.flexiant.extility.Subnet;
import com.flexiant.extility.Vlan;

public class FlexiscaleIP implements IpAddressSupport {
	static private final Logger logger = Logger.getLogger(FlexiscaleIP.class);
	private FlexiscaleProvider provider = null;
	private FlexiScaleSoapBindingStub service = null;	
	
	FlexiscaleIP(FlexiscaleProvider provider) throws CloudException {
		this.provider = provider;
		FlexiscaleStub locator = new FlexiscaleStub (this.provider);
		this.service = locator.getFlexiscaleStub();
	} 
	
	public long getNetworkInterfaceFromServer(String serverId) throws InternalException,	CloudException {
		if(serverId == null)
			return 0L;
		try {
			Server server= service.getServer(Long.valueOf(serverId));			
			if(server != null){
				NetworkInterface[] nics = server.getNetwork_interfaces();
				//Hardcode the nic as the first one
				return nics[0].getNic_id();
			}	
		} catch (NumberFormatException e) {
			logger.error(e.getMessage());
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e.getMessage());
			throw new CloudException(e);
		}
		return 0L;
	}
	
	/** Assign a ip address to a network interface
	 *@param nicId: network interface id 
	 **/
	@Override
	public void assign(String addressId, String instanceId) throws InternalException,	CloudException {
		//Hardcode the nic id as the first nic of each instance
		long nicId = this.getNetworkInterfaceFromServer(instanceId);
		if(nicId != 0L){
			boolean isDhcp = true;
			assign(Long.valueOf(nicId),addressId,isDhcp, instanceId );		
		}
	}
	
	
	public String[] getIPAddressFromNicId(long nicID )throws InternalException,	CloudException {
	
		try {
			NetworkInterface networkInterface= service.getNetworkInterface(nicID);
			return networkInterface.getIp_list();			
		} catch (RemoteException e) {
			logger.error(e.getMessage());
			throw new CloudException(e);
		}	
	}
		
	public void assign(long nicID, String ipAddress, boolean dhcp, String instanceId ) throws InternalException,	CloudException {
		try {
			
			String[] originalAddress = getIPAddressFromNicId(nicID);
			int returncode= service.addIpAddressToNic(nicID, ipAddress, dhcp);
			if(returncode != 1){
				throw new CloudException("Assign Ip "+ ipAddress+ " to the instance was not successful");
			}else{
				// Add the firewall rules for this IP Address with the server's securiyGroup
				
				if(originalAddress == null){
					Collection<String> securityGroupIds = provider.getSecurityGroupIdFromServerId(null, instanceId);
					for(String securityGroupId: securityGroupIds){
						this.provider.addIpAddresseInSecurityGroup(securityGroupId, ipAddress);
					}			
				}			
			}
			
		} catch (NumberFormatException e) {
			logger.error(e.getMessage());
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e.getMessage());
			throw new CloudException(e);
		}
	}

	@Override
    public String forward(String addressId, int publicPort, Protocol protocol, int privatePort, String serverId) throws InternalException, CloudException {
        throw new OperationNotSupportedException("Flexiscale does not support forwarding operation");
    }

	@Override
	public IpAddress getIpAddress(String addressId) throws InternalException, CloudException {
		try {
			String regionId = this.provider.getRegionId();
			if(regionId == null) return null;
			Vlan[] vlans = service.listVLANs(Long.valueOf(regionId));
			Server servers [] = service.listServers(Long.valueOf(regionId));
			if(vlans == null) 
				return null;			
			//Go through each vlan to get the subnet list
			for(Vlan vlan : vlans){
				boolean isPrivate = vlan.get_private();
				Subnet[] subnets = vlan.getSubnets();					
				if(subnets == null) 
					continue;
				
				//Go through each subnet to obtain the ip list
				for(Subnet subnet : subnets){
					String[] ips = this.getIPArrayFromMask(subnet.getNetwork_address(), subnet.getMask());
					if(ips == null) 
						continue;
					
					for(String ip : ips){
						if(ip.trim().equals(addressId)){
							String serverId = getServerId(servers,ip.trim());
							IpAddress address = toAddress(serverId, ip, isPrivate);	
							return address;
						}
					}
				}
			}
	
		} catch (NumberFormatException e) {
			logger.error(e.getMessage());
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e.getMessage());
			throw new CloudException(e);
		}
		return null;
	}

	@Override
	public String getProviderTermForIpAddress(Locale locale) {
		return "Network Address";
	}
	
	/** Get the network interface Id from the ip address
	 *@param
	 **/
	public Long getNetworkInterfaceId(String ipAddress) throws InternalException, CloudException{
		try {
			String regionId = provider.getRegionId();
			Vlan[] vlans = service.listVLANs(Long.valueOf(regionId));
			for(Vlan vlan : vlans){
				NetworkInterface nis [] = service.listNicsForVLAN(vlan.getVlan_id());
				Long nicId = getNetworkInterfaceId(nis,ipAddress);
				if(nicId != null){
					return 	nicId;
				}				
			}
			return null;			
		} catch (NumberFormatException e) {
			logger.error(e.getMessage());
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e.getMessage());
			throw new CloudException(e);
		}
	}
	
	/** Get the network interface Id from the server id and  ip address
	 *@param
	 **/
	public NetworkInterface[] getNetworkInterfaceId(String serverId, String ipAddress) throws InternalException, CloudException{
		try {
			String regionId = this.provider.getRegionId();
			if(regionId == null) return null;

			Server[] servers = service.listServers(Long.valueOf(regionId));
			if(servers == null) return null;
			for(Server server: servers){
				String [] ips = server.getIp_addresses();
				if(ips == null) continue;
				for (String ip : ips){
					if(ip.equals(ipAddress))
						return server.getNetwork_interfaces();
				}
			}
	
			return null;
		} catch (NumberFormatException e) {
			logger.error(e.getMessage());
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e.getMessage());
			throw new CloudException(e);
		}
	}
	
	/**
	 *@param NetworkInterface[] The array of virtual network interface cards within a virtual data center
	 **/
	private Long getNetworkInterfaceId(NetworkInterface[] nis, String ipAddress){
		if(nis == null) return null;
		
		for(NetworkInterface ni : nis){
			String[] ips = ni.getIp_list();			
			if(ips != null){
				for(String ip : ips){
					if(ip.trim().equals(ipAddress)){
						return ni.getNic_id();
					}
				}
			}
		}
		return null;
	}
	/** Get the server Id from the ip address
	 **/
	public String getServerId(String ipAddress) throws InternalException, CloudException{
		try {
			String regionId = this.provider.getRegionId();
			if(regionId == null) return null;
			Server servers [] = service.listServers(Long.valueOf(regionId));
			return getServerId(servers,ipAddress);
	
		} catch (NumberFormatException e) {
			logger.error(e.getMessage());
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e.getMessage());
			throw new CloudException(e);
		}
		
	}
	/** 
	 *@param Server[] The array of virtual machines within a virtual data center
	 **/
	private String getServerId(Server[] servers, String ipAddress){
		if(servers == null) return null;
		
		//Go through each server to obtain the ip list
		for(Server s : servers){
			String[] ips = s.getIp_addresses();			
			if(ips == null) return null;			
			for(String ip : ips){
				if(ip.trim().equals(ipAddress)){
					return String.valueOf(s.getServer_id());
				}
			}
		}
		return null;
	}
	
	
	@Override
	public boolean isAssigned(AddressType type) {
		if(type.equals(AddressType.PUBLIC)){
			return true;			
		}else{
			return false;	
		}			
	}
	
	/**
	 * @param addressId: ip address
	 */		
	public boolean isAssigned(String addressId) throws InternalException, CloudException{
		return (getServerId(addressId)!= null);	
	}

	@Override
	public boolean isForwarding() {
		return false;
	}
	

	@Override
    public boolean isRequestable(AddressType type) {
		if(type.equals(AddressType.PUBLIC)){
			return true;			
		}else{
			return false;	
		}	
    }
	
	public boolean isRequestable(String addressId) throws InternalException, CloudException {
		return !isAssigned(addressId);
	}
	
    @Override
    public boolean isSubscribed() throws CloudException, InternalException {
        return true;
    }
    
	@Override
	public Iterable<IpAddress> listPrivateIpPool(boolean unassignedOnly) throws InternalException, CloudException {
		return listIpPool(false, unassignedOnly);
	}
	private String[] getIPArrayFromMask(String startIp, int mask){
		// The total ipArray should reduce the start network address, gateway, and broadcast
		String[] ipArray = new String[(int) (Math.pow(2, 32-mask) -3) ];
		String prefix = startIp.substring(0,startIp.lastIndexOf(".")+1);
		int suffixValue = Integer.valueOf(startIp.substring(startIp.lastIndexOf(".")+1));
		for(int i = 0; i< 5; i++){
			int suffix = suffixValue + 2 + i;
			ipArray[i] = prefix + suffix;
		}
		return ipArray;
		
	}

	/**
	 * Flexiscale can have multiple VLans for different virtual data centers
	 * */	
	@Override
	public Iterable<IpAddress> listPublicIpPool(boolean unassignedOnly) throws InternalException, CloudException {
		return listIpPool(true, unassignedOnly);
	}
	
	public Iterable<IpAddress> listIpPool(boolean isPublic, boolean unassignedOnly) throws InternalException, CloudException {
		ArrayList<IpAddress> list = new ArrayList<IpAddress>();
		try {
			String regionId = this.provider.getRegionId();
			if(regionId == null) return null;
			
			Vlan[] vlans = service.listVLANs(Long.valueOf(regionId));
			
			if(vlans == null)  return null;
			
			//Go through each vlan to get the subnet list
			for(Vlan vlan : vlans){
				boolean isPrivate = vlan.get_private();
				if(isPublic){
					if(isPrivate){
						continue;
					}									
				}else{
					if(!isPrivate){
						continue;
					}
				}
				Subnet[] subnets = vlan.getSubnets();					
				if(subnets == null) 
					continue;
				
				Server[] servers  = null;
				//Go through each subnet to obtain the ip list
				for(Subnet subnet : subnets){
					String[] ips = null;
					if(unassignedOnly){
						ips = subnet.getAvailable_ip_list();
					}else{
						ips= getIPArrayFromMask(subnet.getNetwork_address(), subnet.getMask());
						//If is unassignedOnly, then need to get the server for each IP
						servers  = service.listServers(Long.valueOf(regionId));
					}
			
					if(ips == null) 
						continue;
					
					for(String ip : ips){
						String serverId = getServerId(servers,ip.trim());
						IpAddress address = toAddress(serverId, ip, isPrivate);	
					    if( address != null && (!unassignedOnly ||(address.getServerId() == null && address.getProviderLoadBalancerId() == null)) ) {
					    	list.add(address);
					    }
					}
				}
			}
	        return list;		
		} catch (NumberFormatException e) {
            logger.error("Virtual data center Id should be long type number " + e.getMessage());
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e.getMessage());
			throw new CloudException(e);
		}
	}


	@Override
	public Collection<IpForwardingRule> listRules(String addressId)	throws InternalException, CloudException {
		return new ArrayList<IpForwardingRule>();
	}

	@Override
	public void releaseFromServer(String addressId) throws InternalException, CloudException {
		Long nicId = getNetworkInterfaceId(addressId);
		if(nicId != null){
			releaseFromServer(nicId, addressId);
		}else{
			logger.error("The address is not associated with a server");
			throw new CloudException("The address is not associated with a server");
		}		
	}	

	private void releaseFromServer(long nicId, String ipAddress) throws InternalException,	CloudException {
		try {
			int returncode= service.removeIpAddressFromNic(nicId, ipAddress);
			if(returncode != 1){
				throw new CloudException("Removing Ip "+ ipAddress+ " from the network interface was not successful");
			}else{
				//Revoke the firewall rules for the ip Address
				this.provider.removeSingleIpAddressInSecurityGroup(ipAddress);
			}
		} catch (NumberFormatException e) {
            logger.error("Network interface Id should be long type number " + e.getMessage());
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e.getMessage());
			throw new CloudException(e);
		}
	}
	
   @Override
   public void releaseFromPool(String addressId) throws InternalException, CloudException {
	   throw new OperationNotSupportedException("Flexiscale does not support to release single IP address, but support to release the subset which includes this address in the FlexiscaleVlan class");
    }
 
   @Override
   public String request(AddressType betterBePublic) throws InternalException, CloudException {
	   try {
		   ArrayList<IpAddress> list = null;
		   boolean isPrivate = false;
		   if(betterBePublic.equals(AddressType.PUBLIC)){
			  list = (ArrayList<IpAddress>) this.listPublicIpPool(true);		
		   }else{
			 // list = (ArrayList<IpAddress>) this.listPrivateIpPool(true);	
			  //isPrivate = true;			   
				logger.error("Does not support to reqeust private IP");
				throw new CloudException("Does not support to reqeust private IP");
		   }	
		   if( list == null){
			   // Can not attach a subnet to a Vlan
			  if(!isPrivate){
				  String inProviderVdcID = this.provider.getRegionId();			
				  Vlan fVlan = service.createVLAN(Long.valueOf(inProviderVdcID), isPrivate);
				  com.flexiant.extility.Subnet fSubnet = service.attachSubnet(fVlan.getVlan_id());
				  if(fSubnet != null){
					  return fSubnet.getAvailable_ip_list()[0];
				  }
			  }
		  }else{
			  if(list.size() > 0)
				  return list.get(0).getProviderIpAddressId();
		  }
		} catch (NumberFormatException e) {
	        logger.error("Network interface Id should be long type number " + e.getMessage());
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e.getMessage());
			throw new CloudException(e);
		}
	   return null;
    }
	   
	@Override
    public void stopForward(String ruleId) throws InternalException, CloudException {
        throw new OperationNotSupportedException("Not support");
    }
	
	private IpAddress toAddress(String instanceId, String ip, boolean isPrivate) throws CloudException {
		IpAddress address = new IpAddress();
		if( ip == null ) {
			throw new CloudException("Invalid address data, no IP.");
		}
		if(isPrivate){
			address.setAddressType(AddressType.PRIVATE);
		}else{
			address.setAddressType(AddressType.PUBLIC);
		}	
		address.setAddress(ip);
		address.setIpAddressId(ip);
		address.setRegionId(provider.getContext().getRegionId());
		address.setServerId(instanceId);		
		return address;
	}

	@Override
	public String[] mapServiceAction(ServiceAction arg0) {
		// TODO Auto-generated method stub
		return new String[0];
	}
}
