/**
 * Copyright (C) 2009-2012 enStratus Networks Inc
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.flexiscale.network;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Locale;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.flexiscale.FlexiscaleProvider;
import org.dasein.cloud.flexiscale.FlexiscaleStub;
import org.dasein.cloud.identity.ServiceAction;
import org.dasein.cloud.network.NetworkInterface;
import org.dasein.cloud.network.Subnet;
import org.dasein.cloud.network.VLAN;
import org.dasein.cloud.network.VLANState;
import org.dasein.cloud.network.VLANSupport;

import com.flexiant.extility.FlexiScaleSoapBindingStub;
import com.flexiant.extility.Server;
import com.flexiant.extility.VDC;
import com.flexiant.extility.Vlan;

public class FlexiscaleVlan implements VLANSupport {
	
	static private final Logger logger = Logger.getLogger(FlexiscaleVlan.class);
	private FlexiscaleProvider provider = null;
	private FlexiScaleSoapBindingStub service = null;	
	
	FlexiscaleVlan(FlexiscaleProvider provider) throws CloudException {
		this.provider = provider;
		FlexiscaleStub locator = new FlexiscaleStub (provider);
		this.service = locator.getFlexiscaleStub();
	} 

	@Override
	public boolean allowsNewSubnetCreation() throws CloudException,InternalException {
		return true;
	}

	@Override
	public boolean allowsNewVlanCreation() throws CloudException,InternalException {
		return true;
	}

	@Override
	public Subnet createSubnet(String cidr, String inProviderVlanId, String name, String description) throws CloudException, InternalException {
		
		// Default null for cidr, name and description
		return createSubnet(inProviderVlanId);
	}
	
	/**
	 * This method is to create a subnet for a specific Vlan
	 * @param inProviderVlanId: flexiscale vlan id
	 * Use null for other arguments 
	 * @return Dasein-Cloud Subnet
	 * */	
	public Subnet createSubnet(String inProviderVlanId) throws CloudException, InternalException {
		try {
			com.flexiant.extility.Subnet fSubnet = service.attachSubnet(Long.valueOf(inProviderVlanId));
			Subnet dSubnet= toSubnet(fSubnet);
			if(dSubnet == null) throw new CloudException("Fail to create create subnet for VLan " + inProviderVlanId );
			return dSubnet;			
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
	}

	@Override
	public VLAN createVlan(String cidr, String name, String description, String domainName, String[] dnsServers, String[] ntpServers) throws CloudException,InternalException {
		//default is public IP
		boolean isPrivate = false;
		String inProviderVdcID = this.provider.getRegionId();
		return createVlan(inProviderVdcID, name, isPrivate);
	}

	/**
	 *This method is to create a Vlan for a specific virtual data center 
	 * @param inProviderVlanId: flexiscale virtual data center id
	 * Use null for other arguments 
	 * @return Dasein-Cloud VLAN
	 *
	 */
	public VLAN createVlan(String inProviderVdcID,String name, boolean isPrivate) throws CloudException,InternalException{
		try {
			Vlan fVlan = service.createVLAN(Long.valueOf(inProviderVdcID), isPrivate);	
			VLAN dVlan = toVLAN(fVlan);
			if(dVlan == null) throw new CloudException("Fail to create Vlan");
			return dVlan;			
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
	}

	@Override
	public int getMaxVlanCount() throws CloudException, InternalException {
		// Not sure yet, need to confirm
		return 4;
	}

	@Override
	public String getProviderTermForNetworkInterface(Locale locale) {
		return "NetworkInterface";
	}

	@Override
	public String getProviderTermForSubnet(Locale locale) {
		return "Subnet";
	}

	@Override
	public String getProviderTermForVlan(Locale locale) {
		return "Vlan";
	}

	@Override
	public Subnet getSubnet(String providerSubnetId) throws CloudException,InternalException {
		
		// Currently, flexiscale has bugs for service.getSubnet(subnetId) method
	/*	try {
			com.flexiant.extility.Subnet fSubnet = service.getSubnet(Long.valueOf(providerSubnetId));
			Subnet dSubnet = toSubnet(fSubnet);
			if( dSubnet != null ) {
				System.out.println("It is: "+ providerSubnetId+"," +dSubnet.getProviderSubnetId());
				return dSubnet;	
			}
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}		
		return null;*/
		
		ArrayList<Subnet> subnets = (ArrayList<Subnet>) listSubnets();
		if(subnets != null){	
			for(int i=0;i < subnets.size();i++){
				String subnetId = subnets.get(i).getProviderSubnetId();
				if(providerSubnetId.equals(subnetId))
					return subnets.get(i);
			}
		}
		return null;
	}

	@Override
	public VLAN getVlan(String providerVlanId) throws CloudException, InternalException {
		try {
			Vlan fVlan = service.getVLAN(Long.valueOf(providerVlanId));
			VLAN dVlan = toVLAN(fVlan);
			if( dVlan != null ) return dVlan;	
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}	
		return null;
	}

	@Override
	public boolean isSubnetDataCenterConstrained() throws CloudException,InternalException {
		// Need confirm
		return true;
	}

	@Override
	public boolean isSubscribed() throws CloudException, InternalException {
		return true;
	}

	@Override
	public boolean isVlanDataCenterConstrained() throws CloudException,InternalException {
		// Need confirm
		return true;
	}

	@Override
	public Iterable<NetworkInterface> listNetworkInterfaces(String providerServerId) throws CloudException, InternalException {
		ArrayList<NetworkInterface> list = new ArrayList<NetworkInterface>();
		try {
			Server fserver = service.getServer(Long.valueOf(providerServerId));
			if(fserver == null) return null;
			com.flexiant.extility.NetworkInterface[]  fnis = fserver.getNetwork_interfaces();
			if(fnis == null) return null;
			for(com.flexiant.extility.NetworkInterface fni: fnis){
				NetworkInterface dni = toNetworkInterface(fni);
				if(dni != null) 
					list.add(dni);
			}
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}	
		return list;
	}
	
	public Iterable<Subnet> listSubnets() throws CloudException, InternalException {
		ArrayList<Subnet> list = new ArrayList<Subnet>();
		try {
			String regionId = provider.getRegionId();			
			com.flexiant.extility.Vlan[] vlans= service.listVLANs(Long.valueOf(regionId));
			for( Vlan vlan : vlans){
			ArrayList<Subnet> newlist = (ArrayList<Subnet>) listSubnets(String.valueOf(vlan.getVlan_id()));
			if(newlist != null) 
				list.addAll(newlist);
			}
			
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}		
		return list;
	}
	

	@Override
	public Iterable<Subnet> listSubnets(String providerVlanId) throws CloudException,InternalException {
		ArrayList<Subnet> list = new ArrayList<Subnet>();
		try {
			com.flexiant.extility.Subnet[] fSubnets = service.listSubnets(Long.valueOf(providerVlanId));
			if(fSubnets == null) return null;
			for(com.flexiant.extility.Subnet fSubnet: fSubnets){
				Subnet dSubnet = toSubnet(fSubnet);
				if(dSubnet != null) 
					list.add(dSubnet);
			}
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}		
		return list;
	}

	@Override
	public Iterable<VLAN> listVlans() throws CloudException, InternalException {
		ArrayList<VLAN> list = new ArrayList<VLAN>();
		try {
			VDC[] vdcs = service.listVDCs();
			if(vdcs == null) return null;
			for(VDC vdc: vdcs){
				ArrayList<VLAN> newlist = (ArrayList<VLAN>) listVlans(String.valueOf(vdc.getVdc_id()));
				if(newlist != null) 
					list.addAll(newlist);
			}
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}		
		return list;
	}
	
	public Iterable<VLAN> listVlans(String virutalDataCenterId) throws CloudException, InternalException {
		ArrayList<VLAN> list = new ArrayList<VLAN>();
		try {
			com.flexiant.extility.Vlan[] fVlans = service.listVLANs(Long.valueOf(virutalDataCenterId));
			if(fVlans == null) return null;
			for(com.flexiant.extility.Vlan fVlan: fVlans){
				VLAN dVlan = toVLAN(fVlan);
				if(dVlan != null) 
					list.add(dVlan);
			}
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}	
		return list;
	}

	@Override
	public void removeSubnet(String providerSubnetId) throws CloudException,InternalException {
		try {
			int returncode = service.removeSubnet(Long.valueOf(providerSubnetId));
			if( returncode != 1 )
				throw new CloudException("Remove the subnet fails");
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}		
	}
	
	@Override
	public void removeVlan(String providerVpcId) throws CloudException,InternalException {
		try {
			int returncode = service.deleteVLAN(Long.valueOf(providerVpcId));
			if( returncode != 1 )
				throw new CloudException("Remove the Vlan fails");
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
	}

	@Override
	public boolean supportsVlansWithSubnets() throws CloudException,InternalException {
		return true;
	}

	private NetworkInterface toNetworkInterface(com.flexiant.extility.NetworkInterface fni ) throws CloudException, InternalException{
		NetworkInterface dni = new NetworkInterface();
		dni.setProviderNetworkInterfaceId(String.valueOf(fni.getNic_id()));
		dni.setProviderVirtualMachineId(String.valueOf(fni.getServer_id()));
		dni.setProviderVlanId(String.valueOf(fni.getVlan_id()));
		
		//Use the first IP Address
		com.flexiant.extility.IpAddress[] fIpAddresss = fni.getIp_details();
		if(fIpAddresss != null){
			com.flexiant.extility.IpAddress fIpAddress = fIpAddresss[0];
			dni.setIpAddress(fIpAddress.getIp_address());
			
			com.flexiant.extility.Subnet subnet = getFlexiscaleSubnet(fIpAddress.getSubnet_id());
			dni.setGatewayAddress(subnet.getGateway_address());
			dni.setDefaultRoute(true);
			dni.setNetmask(toNetworkMask(subnet.getMask()));
		}
		return dni;		
	}
	
	private String toNetworkMask(int mask){
		String prefix = "255.255.255.";
		int value = 256- (int)Math.pow(2.0, 32- mask);
		prefix += value;
		return prefix;		
	}

	private com.flexiant.extility.Subnet getFlexiscaleSubnet(long subnetId)throws CloudException,InternalException{
		try {
			com.flexiant.extility.Subnet subnet = service.getSubnet(subnetId);
			if( subnet != null )
				return subnet;
		} catch (NumberFormatException e) {
			logger.error(e);
			throw new InternalException(e);
		} catch (RemoteException e) {
			logger.error(e);
			throw new CloudException(e);
		}
		return null;
	}
	
	/**
	 * This method is to change the flexiscale Subnet type to dasein-cloud subnet type
	 * @param Vlan: flexicale Subnet 
	 * @return Dasein-Cloud Subnet
	 * */
	private Subnet toSubnet (com.flexiant.extility.Subnet fSubnet)throws CloudException,InternalException{
		Subnet dSubnet = new Subnet ();
		if(fSubnet == null) return null;		
		dSubnet.setProviderOwnerId(provider.getContext().getAccountNumber());
		
		dSubnet.setProviderRegionId(provider.getContext().getRegionId());
		
		dSubnet.setProviderVlanId(String.valueOf(fSubnet.getVlan_id()));
		
		dSubnet.setProviderSubnetId(String.valueOf(fSubnet.getSubnet_id()));
		
		dSubnet.setProviderDataCenterId(provider.getDataCenterId());		
		
		String[] availableIPs= fSubnet.getAvailable_ip_list();
		if(availableIPs != null) 
			dSubnet.setAvailableIpAddresses(availableIPs.length);
				
		// The format of CIDR : IP/mask
		dSubnet.setCidr(fSubnet.getNetwork_address()+ "/" + fSubnet.getMask());
				
		String fSubnetName = dSubnet.getCidr();
		

		dSubnet.setName(fSubnetName);	
		dSubnet.setDescription(fSubnetName);	
		
		HashMap<String,String> tags =  new HashMap<String,String>();		
		tags.put("Gateway", fSubnet.getGateway_address());
		tags.put("Broadcast", fSubnet.getBroadcast_address());
		dSubnet.setTags(tags);
		
		return dSubnet;
	}
		
	/**
	 * This method is to change the flexiscale Vlan type to dasein-cloud VLAN type
	 * @param Vlan: flexiscale Vlan 
	 * @return Dasein-Cloud VLAN
	 * */
	private VLAN toVLAN (Vlan fVlan)throws CloudException,InternalException{
		VLAN dVlan = new VLAN ();
		if(fVlan == null) return null;
		
		//The status is default as available
		dVlan.setCurrentState(VLANState.AVAILABLE);
		
		dVlan.setProviderVlanId(String.valueOf(fVlan.getVlan_id()));
        
		dVlan.setProviderOwnerId(provider.getContext().getAccountNumber());
		dVlan.setProviderRegionId(provider.getContext().getRegionId());
		
		
		dVlan.setProviderDataCenterId(String.valueOf(fVlan.getVdc_id()));
		String dVlanName = fVlan.getVlan_name();
		if(dVlanName == null)
			dVlanName = dVlan.getProviderOwnerId();
		
		// Use the Id as name and description information
		dVlan.setName(dVlanName);
		dVlan.setDescription(dVlan.getName());
		
		HashMap<String,String> tags =  new HashMap<String,String>();
		tags.put("isPrivate", String.valueOf(fVlan.get_private()));
		dVlan.setTags(tags);
		return dVlan;		
	}

	@Override
	public String[] mapServiceAction(ServiceAction arg0) {
		// TODO Auto-generated method stub
		return new String[0];
	}

}
